import networkx as nx
from time import perf_counter

from evaluation.export import write_row
from evaluation.T import (
    GasstationQueryRow,
    ChargeQueryRow,
    ClassicQueryRow,
    AStarQueryRow
)

from evrouting.T import Result
from evrouting import gasstation, charge
from evrouting.graph_tools import (
    consumption_function_distance_factory,
    DISTANCE_KEY
)
from evrouting.osm.profiles import car
from evrouting.osm.routing import GasstationAccessFunctions, a_start_heuristic

__all__ = ['gasstation_query', 'charge_query', 'classic_query', 'astar_query']


def gasstation_query(graph, conf, s, t):
    f = GasstationAccessFunctions(conf['consumption']['consumption_coefficient'])

    start = perf_counter()
    contracted_graph = gasstation.routing.contract_graph(
        G=graph,
        charging_stations=graph.charging_stations,
        capacity=conf['capacity'],
        f=f
    )
    contraction_time = perf_counter() - start

    start = perf_counter()
    state_graph = gasstation.routing.state_graph(
        contracted_graph,
        conf['capacity'],
        f
    )
    state_graph_time = perf_counter() - start

    start = perf_counter()
    result = gasstation.routing.shortest_path(
        G=graph,
        charging_stations=graph.charging_stations,
        s=s,
        t=t,
        initial_soc=conf['mu_s'],
        final_soc=conf['mu_t'],
        capacity=conf['capacity'],
        f=f,
        extended_graph=state_graph,
        contracted_graph=contracted_graph
    )
    query_time = perf_counter() - start

    return GasstationQueryRow(
        start_node=s,
        target_node=t,
        query_time=query_time + contraction_time + state_graph_time,
        trip_time=result.trip_time if type(result) == Result else None,
        nodes=len(graph.nodes),
        edges=len(graph.edges),
        charging_stations=len(graph.charging_stations),
        time_contracted_graph=contraction_time,
        time_state_graph=state_graph_time
    )


def charge_query(graph, conf, s, t):
    start = perf_counter()
    result = charge.shortest_path(
        G=graph,
        charging_stations=graph.charging_stations,
        s=s,
        t=t,
        initial_soc=conf['mu_s'],
        final_soc=conf['mu_t'],
        capacity=conf['capacity'],
        c=consumption_function_distance_factory(conf['consumption']['consumption_coefficient'])
    )
    runtime = start - perf_counter()

    return ChargeQueryRow(
        start_node=s,
        target_node=t,
        query_time=runtime,
        trip_time=result.trip_time if type(result) == Result else None,
        nodes=len(graph.nodes),
        edges=len(graph.edges),
        charging_stations=len(graph.charging_stations)
    )


def classic_query(graph, conf, s, t):
    start = perf_counter()
    try:
        result = nx.shortest_path_length(graph, s, t, weight=DISTANCE_KEY)
    except nx.NetworkXNoPath:
        result = None
    runtime = perf_counter() - start

    return ClassicQueryRow(
        start_node=s,
        target_node=t,
        query_time=runtime,
        trip_time=result,
        nodes=len(graph.nodes),
        edges=len(graph.edges),
        charging_stations=len(graph.charging_stations),
        dijkstra_rank=0
    )


def astar_query(graph, conf, s, t):
    start = perf_counter()
    try:
        result = nx.astar_path_length(
            graph,
            s, t,
            weight=DISTANCE_KEY,
            heuristic=a_start_heuristic(graph, car)
        )
    except nx.NetworkXNoPath:
        result = None
    runtime = perf_counter() - start

    return AStarQueryRow(
        start_node=s,
        target_node=t,
        query_time=runtime,
        trip_time=result,
        nodes=len(graph.nodes),
        edges=len(graph.edges),
        charging_stations=len(graph.charging_stations)
    )
